from typing import List

from yate.protocol import Message

from ywsd.routing_tree import CallTarget, IntermediateRoutingResult


def encode_routing_result(msg: Message, result: IntermediateRoutingResult):
    global_parameters = result.target.parameters
    msg.params.update(global_parameters)
    if result.is_simple:
        msg.return_value = result.target.target
    else:
        msg.return_value = "fork"
        fork_parameters = calltargets_to_callfork_params(
            result.fork_targets, global_parameters
        )
        msg.params.update(fork_parameters)
    return msg


def calltargets_to_callfork_params(
    call_targets: List["CallTarget"], global_params: dict
):
    index = 1
    params = {}
    for target in call_targets:
        current_prefix = "callto." + str(index)
        params[current_prefix] = target.target
        for key, value in target.parameters.items():
            if key not in global_params or global_params[key] != value:
                params[current_prefix + "." + key] = value
        index += 1
    return params



from typing import Optional, Dict
import argparse
import asyncio
import logging
import signal
import traceback

import aiopg.sa
from aiohttp import web

from yate.asyncio import YateAsync
from yate.protocol import Message

import ywsd.yate
from ywsd.objects import Yate, Extension, DoesNotExist
from ywsd import stage1, stage2
from ywsd.util import class_from_dotted_string
from ywsd.routing_cache import RoutingCacheBase
from ywsd.routing_tree import IntermediateRoutingResult, RoutingTree, RoutingError
from ywsd.settings import Settings


class YateRoutingEngine(YateAsync):
    def __init__(self, *args, **kwargs):
        self._settings = kwargs.pop("settings")
        self._web_only = kwargs.pop("web_only")
        super().__init__(*args, **kwargs)
        self._shutdown_future = None
        self._routing_db_engine = None
        self._stage2_db_engine = None
        self._routing_cache: Optional[RoutingCacheBase] = None
        self._yates_dict: Dict[int, Yate] = {}

        if self._settings.WEB_INTERFACE is not None:
            self._web_app = web.Application()
            self._web_app.add_routes([web.get("/stage1", self._web_stage1_handler)])
            self._app_runner = web.AppRunner(self._web_app)
        else:
            self._web_app = None

    @property
    def settings(self):
        return self._settings

    @property
    def yates_dict(self):
        return self._yates_dict

    @property
    def routing_db_engine(self):
        return self._routing_db_engine

    @property
    def stage2_db_engine(self):
        return self._stage2_db_engine

    def run(self):
        if self._web_only:
            logging.info("Starting up in web server only mode")
            self.main_task = self.event_loop.create_task(self.main(42))
            self.event_loop.run_until_complete(self.main_task)
            self.event_loop.close()
        else:
            logging.info("Initializing YateAsync engine")
            super().run(self.main)

    async def main(self, _):
        logging.info("Initializing main application")
        self._shutdown_future = asyncio.get_event_loop().create_future()
        try:
            asyncio.get_event_loop().add_signal_handler(
                signal.SIGINT, lambda: self._shutdown_future.set_result(True)
            )
        except NotImplementedError:
            pass  # Ignore if not implemented. Means this program is running in windows.
        logging.info("Initializing routing cache")
        self._routing_cache = class_from_dotted_string(
            self.settings.CACHE_IMPLEMENTATION
        )(self, self.settings)
        await self._routing_cache.init()

        logging.info("Initializing database engine")
        async with aiopg.sa.create_engine(**self._settings.DB_CONFIG) as db_engine:
            self._routing_db_engine = db_engine
            logging.info("Loading remote yates information from DB")
            async with self.routing_db_engine.acquire() as db_connection:
                self._yates_dict = await Yate.load_yates_dict(db_connection)

            logging.info("Initializing stage2 database engine")
            async with aiopg.sa.create_engine(
                **self._settings.STAGE2_DB_CONFIG
            ) as stage2_db_engine:
                self._stage2_db_engine = stage2_db_engine

                # fire up http server if requested
                if self._web_app is not None:
                    bind = self._settings.WEB_INTERFACE.get("bind_address")
                    port = int(self._settings.WEB_INTERFACE.get("port", 9000))
                    await self._app_runner.setup()
                    site = web.TCPSite(self._app_runner, bind, port)
                    await site.start()
                    logging.info(
                        "Webserver ready. Waiting for requests on {}:{}.".format(
                            bind, port
                        )
                    )

                if not self._web_only:
                    logging.info("Registering for routing messages")
                    if not await self.register_message_handler_async(
                        "call.route", self._call_route_handler, 90
                    ):
                        logging.error("Cannot register for call.route. Terminating...")
                        return
                    await self.activate_automatic_bufsize()
                    logging.info(
                        "Yate buffer size defaults to {}".format(
                            self.get_local("bufsize")
                        )
                    )

                logging.info("Ready to route")
                await self._shutdown_future

        await self._routing_cache.stop()
        self._routing_db_engine = None
        self._stage2_db_engine = None

    def _call_route_handler(self, msg: Message) -> Optional[bool]:
        logging.debug("Asked to route message: {}".format(msg.params))
        called = msg.params.get("called")
        stage2_active = msg.params.get("eventphone_stage2", "0")

        if called is None or called == "":
            return False
        if called.isdigit():
            if (
                msg.params.get("connection_id", "")
                == self.settings.INTERNAL_YATE_LISTENER
                or stage2_active == "1"
            ):
                task = stage2.RoutingTask(self, msg)
            else:
                task = stage1.RoutingTask(self, msg)
            self.event_loop.create_task(task.routing_job())
        elif called.startswith("stage1-"):
            self.event_loop.create_task(self._retrieve_from_cache_for(msg))
        elif called.startswith("stage2-"):
            task = stage2.RoutingTask(self, msg)
            self.event_loop.create_task(task.routing_job())
        else:
            return False

    async def _retrieve_from_cache_for(self, msg: Message):
        called = "lateroute/" + msg.params.get("called")
        result = await self._routing_cache.retrieve(called)
        if result is None:
            # This is an invalid entry, answer the message but with invalid result
            msg.result = ""
            self.answer_message(msg, True)
        else:
            msg = ywsd.yate.encode_routing_result(msg, result)
            self.answer_message(msg, True)

    async def store_cache_infos(self, entries: Dict[str, IntermediateRoutingResult]):
        await self._routing_cache.update(entries)

    async def _web_stage1_handler(self, request):
        params = request.query

        called = params.get("called")
        caller = params.get("caller")

        if any((caller is None, called is None)):
            return web.Response(
                status=400, text="Provide at least <caller> and <called>"
            )

        # calculate routing tree and results
        routing_tree = None
        routing_result = None
        all_routing_results = None
        routing_cache_entries = {}
        routing_status = "PROCESSING"
        routing_status_details = ""
        try:
            async with self.routing_db_engine.acquire() as db_connection:
                try:
                    caller_extension = await Extension.load_extension(
                        caller, db_connection
                    )
                except DoesNotExist:
                    caller_extension = Extension.create_external(caller)
                caller_params = stage1.RoutingTask.calculate_source_parameters(
                    caller_extension
                )
                routing_tree = RoutingTree(
                    caller_extension, called, caller_params, self.settings
                )
                await routing_tree.discover_tree(db_connection)

            routing_result, routing_cache_entries = routing_tree.calculate_routing(
                self.settings.LOCAL_YATE_ID, self.yates_dict
            )
            all_routing_results = routing_tree.all_routing_results
            routing_status = "OK"
        except RoutingError as e:
            routing_status = "ERROR"
            all_routing_results = {}
            routing_status_details = "{}: {}".format(e.error_code, e.message)
        except Exception as e:
            backtrace = traceback.format_exc()
            routing_status = "ERROR"
            all_routing_results = {}
            routing_status_details = (
                "Unexpected Exception while routing:\n{}:Backtrace:\n{}".format(
                    e, backtrace
                )
            )

        json_response_data = {
            "routing_tree": routing_tree.serialized_tree(),
            "main_routing_result": routing_result.serialize()
            if routing_result is not None
            else None,
            "all_routing_results": {
                key: result.serialize()
                for key, result in all_routing_results.items()
                if result.is_valid and result.target.target in routing_cache_entries
            },
            "routing_status": routing_status,
            "routing_status_details": routing_status_details,
        }
        return web.json_response(json_response_data)


def main():
    parser = argparse.ArgumentParser(description="Yate Routing Engine")
    parser.add_argument(
        "--config", type=str, help="Config file to use.", default="routing_engine.yaml"
    )
    parser.add_argument("--verbose", help="Print out debug logs.", action="store_true")
    parser.add_argument(
        "--web-only",
        help="Only start the webserver. Do not connect to yate",
        dest="web_only",
        action="store_true",
    )

    args = parser.parse_args()
    settings = Settings(args.config)

    logging_basic_config_params = {
        "format": "%(asctime)s:%(name)-10s:%(levelname)-8s:%(message)s",
        "datefmt": "%Y-%m-%d:%H:%M:%S",
    }
    if settings.LOG_FILE is not None:
        logging_basic_config_params["filename"] = settings.LOG_FILE
        logging_basic_config_params["filemode"] = "a+"

    if args.verbose or settings.LOG_VERBOSE:
        logging.basicConfig(level=logging.DEBUG, **logging_basic_config_params)
    else:
        logging.basicConfig(level=logging.INFO, **logging_basic_config_params)

    logging.debug("Debug logging enabled.")

    yate_connection = settings.YATE_CONNECTION
    app = YateRoutingEngine(
        settings=settings, web_only=args.web_only, **yate_connection
    )
    app.run()


if __name__ == "__main__":
    main()



import asyncio

from ywsd.settings import Settings
from ywsd.objects import *
from aiopg.sa import create_engine

from ywsd.routing_tree import RoutingTree


settings = Settings()


async def reinstall_testdata():
    async with create_engine(**settings.DB_CONFIG) as engine:
        async with engine.acquire() as conn:
            await regenerate_database_objects(conn)
            await write_testdata(conn)


async def get_extension(ext):
    async with create_engine(**settings.DB_CONFIG) as engine:
        async with engine.acquire() as conn:
            res = await Extension.load_extension(ext, conn)
            return res


async def exec_with_db(func):
    async with create_engine(**settings.DB_CONFIG) as engine:
        async with engine.acquire() as conn:
            return await func(conn)


async def async_test_route(src, target, local_yate):
    async with create_engine(**settings.DB_CONFIG) as engine:
        async with engine.acquire() as conn:
            yates = await Yate.load_yates_dict(conn)
            tree = RoutingTree(src, target, settings)
            await tree.discover_tree(conn)
            tree.calculate_routing(local_yate, yates)
            return tree


def test_route(src, target, local_yate):
    l = asyncio.get_event_loop()
    return l.run_until_complete(async_test_route(src, target, local_yate))


def test_route_display(src, target, local_yate):
    tree = test_route(src, target, local_yate)
    print(repr(tree.routing_result))
    print(repr(tree.new_routing_cache_content))


yates_dict = {1: "dect", 2: "sip", 3: "app"}


async def write_testdata(conn):
    await conn.execute(
        Yate.table.insert().values(
            [
                {
                    "hostname": "dect",
                    "voip_listener": "local",
                    "guru3_identifier": "DECT",
                },
                {
                    "hostname": "sip",
                    "voip_listener": "local",
                    "guru3_identifier": "SIP",
                },
                {
                    "hostname": "app",
                    "voip_listener": "local",
                    "guru3_identifier": "APP",
                },
            ]
        )
    )

    yates = {}
    async for row in conn.execute(Yate.table.select()):
        yates[row.hostname] = row.id

    print(repr(yates))

    await conn.execute(
        Extension.table.insert().values(
            [
                {
                    "yate_id": None,
                    "extension": "2000",
                    "name": "PoC",
                    "type": "GROUP",
                    "forwarding_mode": "DISABLED",
                    "lang": "de_DE",
                },
                {
                    "yate_id": yates["dect"],
                    "extension": "2001",
                    "name": "PoC Sascha",
                    "type": "MULTIRING",
                    "forwarding_mode": "DISABLED",
                    "lang": "de_DE",
                },
                {
                    "yate_id": yates["dect"],
                    "extension": "2002",
                    "name": "PoC Bernie",
                    "type": "SIMPLE",
                    "forwarding_mode": "DISABLED",
                    "lang": "de_DE",
                },
                {
                    "yate_id": yates["dect"],
                    "extension": "2004",
                    "name": "PoC BeF",
                    "type": "SIMPLE",
                    "forwarding_mode": "DISABLED",
                    "lang": "de_DE",
                },
                {
                    "yate_id": yates["sip"],
                    "extension": "2005",
                    "name": "PoC Sascha (SIP)",
                    "type": "SIMPLE",
                    "forwarding_mode": "DISABLED",
                    "lang": "de_DE",
                },
                {
                    "yate_id": yates["sip"],
                    "extension": "2042",
                    "name": "PoC Garwin",
                    "type": "SIMPLE",
                    "forwarding_mode": "DISABLED",
                    "lang": "de_DE",
                },
            ]
        )
    )

    exts = {}
    async for row in conn.execute(Extension.table.select()):
        exts[row.extension] = row.id
    print(repr(exts))

    await conn.execute(
        ForkRank.table.insert().values(
            [
                {"extension_id": exts["2000"], "index": 0, "mode": "DEFAULT"},
                {"extension_id": exts["2001"], "index": 0, "mode": "DEFAULT"},
            ]
        )
    )

    cgr = {}
    async for row in conn.execute(ForkRank.table.select()):
        cgr[row.extension_id] = row.id
    print(repr(cgr))

    await conn.execute(
        ForkRank.member_table.insert().values(
            [
                {
                    "forkrank_id": cgr[exts["2000"]],
                    "extension_id": exts["2001"],
                    "rankmember_type": "DEFAULT",
                    "active": True,
                },
                {
                    "forkrank_id": cgr[exts["2000"]],
                    "extension_id": exts["2002"],
                    "rankmember_type": "DEFAULT",
                    "active": True,
                },
                {
                    "forkrank_id": cgr[exts["2000"]],
                    "extension_id": exts["2004"],
                    "rankmember_type": "DEFAULT",
                    "active": True,
                },
                {
                    "forkrank_id": cgr[exts["2000"]],
                    "extension_id": exts["2042"],
                    "rankmember_type": "DEFAULT",
                    "active": True,
                },
                {
                    "forkrank_id": cgr[exts["2001"]],
                    "extension_id": exts["2005"],
                    "rankmember_type": "DEFAULT",
                    "active": True,
                },
            ]
        )
    )



import json
import logging

from typing import Dict, Optional

from ywsd.routing_tree import IntermediateRoutingResult


class CacheError(Exception):
    def __init__(self, error_string):
        self._error_string = error_string

    def __str__(self):
        return self._error_string


class RoutingCacheBase:
    def __init__(self, yate, settings):
        pass

    async def init(self):
        pass

    async def stop(self):
        pass

    async def retrieve(self, target) -> Optional[IntermediateRoutingResult]:
        pass

    async def update(self, results: Dict[str, IntermediateRoutingResult]):
        pass


class PythonDictRoutingCache(RoutingCacheBase):
    def __init__(self, yate, settings):
        self._cache = {}

    async def retrieve(self, target) -> Optional[IntermediateRoutingResult]:
        return self._cache.get(target)

    async def update(self, results: Dict[str, IntermediateRoutingResult]):
        self._cache.update(results)


# If aioredis is not available, this should not be a failure. We just don't support this backend then
try:
    import aioredis
except ImportError:
    pass


class RedisRoutingCache(RoutingCacheBase):
    def __init__(self, yate, settings):
        self._address = settings.CACHE_CONFIG.get("address")
        if self._address is None:
            raise CacheError("No address configured for redis.")

        self._object_lifetime = settings.CACHE_CONFIG.get("object_lifetime", 600)

        self._redis = None

    async def init(self):
        try:
            self._redis = await aioredis.create_redis_pool(self._address, timeout=20)
            logging.info("Conected to redis routing cache.")
        except (FileNotFoundError, OSError) as e:
            raise CacheError("Unable to connect to redis: {}".format(e))

    async def stop(self):
        if self._redis is not None:
            self._redis.close()
            await self._redis.wait_closed()

    async def retrieve(self, target) -> Optional[IntermediateRoutingResult]:
        data = await self._redis.get(target)
        if data is None:
            return None
        data = json.loads(data)
        return IntermediateRoutingResult.deserialize(data)

    async def update(self, results: Dict[str, IntermediateRoutingResult]):
        for key, routing_result in results.items():
            await self._redis.set(
                key,
                json.dumps(routing_result.serialize()),
                expire=self._object_lifetime,
            )



import logging
import traceback

from yate.protocol import Message

import ywsd.yate
from ywsd.objects import Extension, DoesNotExist
from ywsd.routing_tree import RoutingTree, RoutingError
from ywsd.util import retry_db_offline, OperationalError


class RoutingTask:
    def __init__(self, yate: "ywsd.engine.YateRoutingEngine", message: Message):
        self._yate = yate
        self._message = message

    async def _sanitize_caller(self, caller, db_connection) -> Extension:
        # if it comes from the internal yate listener, we just trust it
        if (
            self._message.params.get("connection_id", "")
            == self._yate.settings.INTERNAL_YATE_LISTENER
        ):
            return caller
        else:
            try:
                caller_extension = await Extension.load_extension(caller, db_connection)
                username = self._message.params.get("username")
                if username is None:
                    raise RoutingError("noauth", "User needs authentication")
                if username != caller:
                    logging.warning(
                        "User {} tries to act as caller {}. Returned Deny.".format(
                            username, caller
                        )
                    )
                    raise RoutingError(
                        "forbidden", "Invalid authentication for this caller"
                    )
                return caller_extension
            except DoesNotExist:
                # this caller doesn't exist in our database, create an external extension
                return Extension.create_external(caller)

    @staticmethod
    def calculate_source_parameters(source: Extension):
        # push parameters here like faked-caller-id or caller-language
        source_parameters = {}
        if source.outgoing_extension is not None and source.outgoing_extension != "":
            source_parameters["caller"] = source.outgoing_extension
            source_parameters["callername"] = source.outgoing_name
        else:
            # avoid name spoofing
            source_parameters["callername"] = source.name
        if source.lang is not None:
            source_parameters["osip_X-Caller-Language"] = source.lang
        if source.dialout_allowed:
            source_parameters["osip_X-Dialout-Allowed"] = "1"
        return source_parameters

    async def _calculate_stage1_routing(self, caller, called):
        try:
            async with self._yate.routing_db_engine.acquire() as db_connection:
                caller = await self._sanitize_caller(caller, db_connection)
                if caller.type != Extension.Type.EXTERNAL:
                    caller_params = RoutingTask.calculate_source_parameters(caller)
                else:
                    caller_params = {}

                logging.debug("Routing {} to {}".format(caller, called))
                routing_tree = RoutingTree(
                    caller, called, caller_params, self._yate.settings
                )
                await routing_tree.discover_tree(db_connection)

            routing_result, routing_cache_entries = routing_tree.calculate_routing(
                self._yate.settings.LOCAL_YATE_ID, self._yate.yates_dict
            )
            logging.debug(
                "Routing result:\n{}\n{}".format(routing_result, routing_cache_entries)
            )

            await self._yate.store_cache_infos(routing_cache_entries)
            return ywsd.yate.encode_routing_result(self._message, routing_result), True
        except RoutingError as e:
            if e.error_code != "noroute":
                self._message.params["error"] = e.error_code
                logging.info(
                    "Routing {} to {} failed: {}".format(caller, called, e.message)
                )
                return self._message, True
            else:
                # We decided that we do not handle the noroute case and give others (regexroute) a chance but
                # populate the caller parameters
                logging.debug(
                    "Routing {} to {} returned noroute, populate caller params and pass on".format(
                        caller, called
                    )
                )
                self._message.params.update(caller_params)
                return self._message, False
        except Exception as e:
            if type(e) == OperationalError:
                raise  # this is a database error and the routing will be re-tried
            backtrace = traceback.format_exc()
            logging.error(
                "An error occurred while routing {} to {}: {}\nBacktrace:".format(
                    caller, called, e, backtrace
                )
            )
            self._message.params["error"] = "failure"
            return self._message, True

    @retry_db_offline(count=4, wait_ms=1000)
    async def routing_job(self):
        caller = self._message.params.get("caller")
        called = self._message.params.get("called")
        if caller is None:
            # we do not process messages without a caller
            self._yate.answer_message(self._message, False)
        result_message, handled = await self._calculate_stage1_routing(caller, called)
        self._yate.answer_message(result_message, handled)



import asyncio
from collections import namedtuple
from enum import Enum

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import ENUM
from sqlalchemy.sql.ddl import CreateTable, CreateIndex
from sqlalchemy.sql.expression import bindparam

metadata = sa.MetaData()


def _plain_loader(fields, source, target, prefix=None):
    for field in fields:
        setattr(
            target, field, getattr(source, field if prefix is None else prefix + field)
        )


def _transform_loader(fields, source, target, prefix=None):
    for field, transform in fields:
        setattr(
            target,
            field,
            transform(getattr(source, field if prefix is None else prefix + field)),
        )


class DoesNotExist(Exception):
    pass


class User:
    table = sa.Table(
        "users",
        metadata,
        sa.Column(
            "username", sa.String(32), unique=True, nullable=False, primary_key=True
        ),
        sa.Column(
            "displayname",
            sa.String(64),
            server_default="EventphoneUser",
            nullable=False,
        ),
        sa.Column("password", sa.String(128), nullable=False),
        sa.Column("inuse", sa.Integer, nullable=False, server_default="0"),
        sa.Column("type", sa.String(20), server_default="user"),
        sa.Column(
            "dect_displaymode",
            ENUM("NUMBER", "NUMBER_AND_NAME", "NAME", name="dect_displaymode"),
        ),
        sa.Column("trunk", sa.Boolean, nullable=False, server_default="0"),
        sa.Column("call_waiting", sa.Boolean, nullable=False, server_default="1"),
        sa.Column("static_target", sa.String(1024), nullable=False, server_default=""),
        sa.CheckConstraint(
            "(password != '') OR (type != 'user')",
            name="sip_users_have_password",
        ),
    )

    FIELDS_PLAIN = (
        "username",
        "displayname",
        "password",
        "inuse",
        "type",
        "trunk",
        "call_waiting",
        "static_target",
    )
    FIELDS_TRANSFORM = (
        (
            "dect_displaymode",
            lambda x: User.DectDisplaymode[x] if x is not None else None,
        ),
    )

    class DectDisplaymode(Enum):
        NUMBER = 0
        NUMBER_AND_NAME = 1
        NAME = 3

    def __init__(self, db_row, prefix=None):
        _plain_loader(self.FIELDS_PLAIN, db_row, self, prefix=prefix)
        _transform_loader(self.FIELDS_TRANSFORM, db_row, self, prefix=prefix)

    @classmethod
    async def load_user(cls, username, db_connection):
        res = await db_connection.execute(
            cls.table.select().where(cls.table.c.username == username)
        )
        if res.rowcount == 0:
            raise DoesNotExist('No user "{}" found'.format(username))
        return cls(await res.first())

    @classmethod
    async def load_trunk(cls, dialed_number, db_connection):
        res = await db_connection.execute(
            cls.table.select()
            .where(
                bindparam("dialed_number", dialed_number).startswith(
                    cls.table.c.username
                )
            )
            .where(cls.table.c.trunk == True)
        )
        if res.rowcount == 0:
            raise DoesNotExist('No trunk for "{}" found'.format(dialed_number))
        elif res.rowcount > 1:
            raise DoesNotExist(
                "Trunk misconfiguration lead to multiple results for {}".format(
                    dialed_number
                )
            )
        return cls(await res.first())


class Registration:
    table = sa.Table(
        "registrations",
        metadata,
        sa.Column(
            "username",
            sa.String(32),
            sa.ForeignKey("users.username", ondelete="CASCADE"),
            nullable=False,
        ),
        sa.Column("location", sa.String(1024), nullable=False),
        sa.Column("oconnection_id", sa.String(1024), nullable=False),
        sa.Column("expires", sa.TIMESTAMP, nullable=False),
        sa.UniqueConstraint(
            "username", "location", "oconnection_id", name="uniq_registrations"
        ),
    )

    FIELDS_PLAIN = ("username", "location", "oconnection_id", "expires")

    def __init__(self, db_row, prefix=None, user=None, dialed_number=None):
        _plain_loader(self.FIELDS_PLAIN, db_row, self, prefix=prefix)
        self._user = user
        self._dialed_number = dialed_number

    @classmethod
    async def load_locations_for(cls, user: User, dialed_number, db_connection):
        result = []
        res = await db_connection.execute(
            cls.table.select().where(cls.table.c.username == user.username)
        )
        return [cls(row, user=user, dialed_number=dialed_number) async for row in res]

    @property
    def call_target(self):
        if self._user is None or self._user.trunk == False:
            return self.location
        # the location field has the format sip/sip:<user>@<ip>:<port>;<param>=<val>,...
        # for a trunk we want to exchange user by the actually dialed number

        return self.location.replace(
            f"{self._user.username}@", f"{self._dialed_number}@", 1
        )


class ActiveCall:
    table = sa.Table(
        "active_calls",
        metadata,
        sa.Column(
            "username",
            sa.String(32),
            sa.ForeignKey("users.username", ondelete="CASCADE"),
            nullable=False,
        ),
        sa.Column("x_eventphone_id", sa.String(64), nullable=False),
    )

    @classmethod
    async def is_active_call(cls, username, x_eventphone_id, db_connection):
        return (
            await db_connection.scalar(
                cls.table.count()
                .where(cls.table.c.username == username)
                .where(cls.table.c.x_eventphone_id == x_eventphone_id)
            )
        ) > 0


class Yate:
    table = sa.Table(
        "Yate",
        metadata,
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("hostname", sa.String(256), nullable=False),
        sa.Column("guru3_identifier", sa.String(32), nullable=False),
        sa.Column("voip_listener", sa.String(256), nullable=False),
    )
    FIELDS_PLAIN = ("id", "hostname", "guru3_identifier", "voip_listener")

    def __init__(self, db_row):
        _plain_loader(self.FIELDS_PLAIN, db_row, self)

    @classmethod
    async def load_yates_dict(cls, db_connection):
        yates_dict = {}
        result = await db_connection.execute(cls.table.select())
        async for row in result:
            yates_dict[row.id] = cls(row)
        return yates_dict


class RoutingTreeNode:
    class LogEntry:
        def __init__(self, msg, level, related_node=None):
            self.msg = msg
            self.level = level
            self.related_node = related_node

        def serialize(self):
            return {
                "msg": self.msg,
                "level": self.level,
                "related_node": self.related_node.tree_identifier,
            }

    def __init__(self):
        self._log = []
        self._tree_identifier = None

    @property
    def tree_identifier(self):
        return self._tree_identifier

    @tree_identifier.setter
    def tree_identifier(self, ti):
        self._tree_identifier = ti

    def routing_log(self, msg, level, related_node=None):
        self._log.append(RoutingTreeNode.LogEntry(msg, level, related_node))

    def serialize(self):
        data = {key: getattr(self, key) for key in self.FIELDS_PLAIN}
        data.update({key: str(getattr(self, key)) for key, _ in self.FIELDS_TRANSFORM})
        data["tree_identifier"] = self.tree_identifier
        data["logs"] = [entry.serialize() for entry in self._log]
        return data


class Extension(RoutingTreeNode):
    table = sa.Table(
        "Extension",
        metadata,
        sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
        sa.Column("yate_id", sa.Integer, sa.ForeignKey("Yate.id")),
        sa.Column("extension", sa.String(32), nullable=False, unique=True),
        sa.Column("name", sa.String(64)),
        sa.Column("short_name", sa.String(8)),
        sa.Column(
            "type",
            ENUM(
                "SIMPLE",
                "MULTIRING",
                "GROUP",
                "EXTERNAL",
                "TRUNK",
                name="extension_type",
            ),
            nullable=False,
        ),
        sa.Column("outgoing_extension", sa.String(32)),
        sa.Column("outgoing_name", sa.String(64)),
        sa.Column("dialout_allowed", sa.Boolean, server_default="0"),
        sa.Column("ringback", sa.String(128)),
        sa.Column(
            "forwarding_mode",
            ENUM(
                "DISABLED",
                "ENABLED",
                "ON_BUSY",
                "ON_UNAVAILABLE",
                name="forwarding_mode",
            ),
            nullable=False,
        ),
        sa.Column("forwarding_delay", sa.Integer),
        sa.Column(
            "forwarding_extension_id",
            sa.Integer,
            sa.ForeignKey("Extension.id", ondelete="SET NULL"),
        ),
        sa.Column("lang", sa.String(6), nullable=False),
        sa.CheckConstraint(
            "(forwarding_extension_id IS NOT NULL) OR (forwarding_mode = 'DISABLED')",
            name="fwd_correct",
        ),
        sa.CheckConstraint(
            "(forwarding_mode != 'ENABLED') OR (forwarding_delay IS NOT NULL)",
            name="fwd_delay_correct",
        ),
        sa.CheckConstraint(
            "(yate_id IS NOT NULL) OR (type != 'SIMPLE' AND type != 'MULTIRING')",
            name="yate_id_not_null_for_direct_ring",
        ),
    )
    FIELDS_PLAIN = (
        "id",
        "yate_id",
        "extension",
        "name",
        "short_name",
        "outgoing_extension",
        "outgoing_name",
        "dialout_allowed",
        "ringback",
        "forwarding_delay",
        "forwarding_extension_id",
        "lang",
    )
    FIELDS_TRANSFORM = (
        ("type", lambda x: Extension.Type[x]),
        ("forwarding_mode", lambda x: Extension.ForwardingMode[x]),
    )

    class Type(Enum):
        SIMPLE = 0
        MULTIRING = 1
        GROUP = 2
        EXTERNAL = 3
        TRUNK = 4

    class ForwardingMode(Enum):
        DISABLED = 0
        ENABLED = 1
        ON_BUSY = 2
        ON_UNAVAILABLE = 3

    def __init__(self, db_row, prefix=None):
        super().__init__()
        _plain_loader(self.FIELDS_PLAIN, db_row, self, prefix=prefix)
        _transform_loader(self.FIELDS_TRANSFORM, db_row, self, prefix=prefix)

        self.fork_ranks = []
        self.forwarding_extension = None

    def serialize(self):
        data = super().serialize()
        if self.forwarding_extension is not None:
            data["forwarding_extension"] = self.forwarding_extension.serialize()
        if self.fork_ranks:
            data["fork_ranks"] = [rank.serialize() for rank in self.fork_ranks]
        return data

    def __repr__(self):
        return "<Extension {}, name={}, type={}>".format(
            self.extension, self.name, self.type
        )

    @classmethod
    def create_external(cls, extension, external_name=None):
        if external_name is None:
            external_name = "External"
        params = {
            "id": None,
            "yate_id": None,
            "extension": extension,
            "name": external_name,
            "short_name": None,
            "outgoing_extension": None,
            "outgoing_name": None,
            "dialout_allowed": False,
            "ringback": None,
            "forwarding_delay": None,
            "forwarding_extension_id": None,
            "lang": None,
            "type": "EXTERNAL",
            "forwarding_mode": "DISABLED",
        }
        return cls(namedtuple("Ext_row", params.keys())(*params.values()))

    @classmethod
    def create_unknown(cls, extension):
        params = {
            "id": None,
            "yate_id": None,
            "extension": extension,
            "name": "Unknown",
            "short_name": None,
            "outgoing_extension": None,
            "outgoing_name": None,
            "dialout_allowed": False,
            "ringback": None,
            "forwarding_delay": None,
            "forwarding_extension_id": None,
            "lang": None,
            "type": "SIMPLE",
            "forwarding_mode": "DISABLED",
        }
        return cls(namedtuple("Ext_row", params.keys())(*params.values()))

    @classmethod
    async def load_extension(cls, extension, db_connection):
        res = await db_connection.execute(
            cls.table.select().where(cls.table.c.extension == extension)
        )
        if res.rowcount == 0:
            raise DoesNotExist('No extension "{}" found'.format(extension))
        return cls(await res.first())

    @classmethod
    async def load_trunk_extension(cls, dialed_number, db_connection):
        res = await db_connection.execute(
            cls.table.select()
            .where(
                bindparam("dialed_number", dialed_number).startswith(
                    cls.table.c.extension
                )
            )
            .where(cls.table.c.type == "TRUNK")
        )
        if res.rowcount == 0:
            raise DoesNotExist('No trunk for "{}" found'.format(dialed_number))
        elif res.rowcount > 1:
            raise DoesNotExist(
                "Trunk misconfiguration lead to multiple results for {}".format(
                    dialed_number
                )
            )
        return cls(await res.first())

    async def load_forwarding_extension(self, db_connection):
        if self.forwarding_extension_id is None:
            raise DoesNotExist("This extension has no forwarding extension")
        res = await db_connection.execute(
            self.table.select().where(self.table.c.id == self.forwarding_extension_id)
        )
        # this always exists and is unique by db constraints
        self.forwarding_extension = Extension(await res.first())
        if self.tree_identifier is not None:
            self.forwarding_extension.tree_identifier = (
                self.tree_identifier + "-" + str(self.forwarding_extension.id)
            )

    async def populate_fork_ranks(self, db_connection):
        result = await db_connection.execute(
            sa.select(
                [ForkRank.table, ForkRank.member_table, Extension.table],
                use_labels=True,
            )
            .where(ForkRank.table.c.extension_id == self.id)
            .where(ForkRank.member_table.c.extension_id == Extension.table.c.id)
            .where(ForkRank.table.c.id == ForkRank.member_table.c.forkrank_id)
            .order_by(ForkRank.table.c.index)
        )
        self.fork_ranks = []
        current_rank_id = None
        current_rank = None
        async for row in result:
            if current_rank_id != row.ForkRank_id:
                current_rank_id = row.ForkRank_id
                current_rank = ForkRank(row, prefix="ForkRank_")
                if self.tree_identifier is not None:
                    current_rank.tree_identifier = (
                        self.tree_identifier + "-fr" + str(current_rank.id)
                    )
                self.fork_ranks.append(current_rank)
            member = ForkRank.Member(
                ForkRank.RankMemberType[row.ForkRankMember_rankmember_type],
                row.ForkRankMember_active,
                Extension(row, prefix="Extension_"),
            )
            if self.tree_identifier is not None:
                member.extension.tree_identifier = (
                    current_rank.tree_identifier + "-" + str(member.extension.id)
                )
            current_rank.members.append(member)

    @property
    def immediate_forward(self):
        return (
            self.forwarding_mode == Extension.ForwardingMode.ENABLED
            and self.forwarding_delay == 0
        )

    @property
    def has_active_group_members(self):
        for rank in self.fork_ranks:
            if any([m.active for m in rank.members]):
                return True
        return False


class ForkRank(RoutingTreeNode):
    table = sa.Table(
        "ForkRank",
        metadata,
        sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
        sa.Column(
            "extension_id",
            sa.Integer,
            sa.ForeignKey("Extension.id", ondelete="CASCADE"),
            nullable=False,
            index=True,
        ),
        sa.Column("index", sa.Integer, nullable=False),
        sa.Column(
            "mode",
            ENUM("DEFAULT", "NEXT", "DROP", name="fork_rank_mode"),
            nullable=False,
        ),
        sa.Column("delay", sa.Integer),
        sa.CheckConstraint(
            "(mode = 'DEFAULT') OR (delay IS NOT NULL)", name="delay_correct"
        ),
    )
    FIELDS_PLAIN = ("id", "extension_id", "index", "delay")
    FIELDS_TRANSFORM = (("mode", lambda x: ForkRank.Mode[x]),)
    member_table = sa.Table(
        "ForkRankMember",
        metadata,
        sa.Column(
            "forkrank_id",
            sa.Integer,
            sa.ForeignKey("ForkRank.id", ondelete="CASCADE"),
            nullable=False,
            index=True,
        ),
        sa.Column(
            "extension_id",
            sa.Integer,
            sa.ForeignKey("Extension.id", ondelete="CASCADE"),
            nullable=False,
        ),
        sa.Column(
            "rankmember_type",
            ENUM("DEFAULT", "AUXILIARY", "PERSISTENT", name="fork_rankmember_type"),
            nullable=False,
        ),
        sa.Column("active", sa.Boolean, nullable=False),
        sa.UniqueConstraint("forkrank_id", "extension_id", name="uniq1"),
    )

    class Mode(Enum):
        DEFAULT = 0
        NEXT = 1
        DROP = 2

    class RankMemberType(Enum):
        DEFAULT = 0
        AUXILIARY = 1
        PERSISTENT = 2

        @property
        def is_special_calltype(self):
            return self != ForkRank.RankMemberType.DEFAULT

        @property
        def fork_calltype(self):
            return self.name.lower()

    class Member:
        def __init__(self, type, active, extension):
            self.type = type
            self.active = active
            self.extension = extension

        def __repr__(self):
            res = "<({})Member: {}".format(self.type, self.extension)
            if self.active:
                return res + ">"
            else:
                return res + " (inactive)>"

    def __init__(self, db_row, prefix=None):
        super().__init__()
        _plain_loader(self.FIELDS_PLAIN, db_row, self, prefix=prefix)
        _transform_loader(self.FIELDS_TRANSFORM, db_row, self, prefix=prefix)
        self.members = []

    def __repr__(self):
        return "<ForkRank id={}, extension_id={}, index={}, mode={}, delay={}>".format(
            self.id, self.extension_id, self.mode, self.index, self.mode, self.delay
        )

    def serialize(self):
        data = super().serialize()
        data["members"] = [
            {
                "type": str(member.type),
                "active": member.active,
                "extension": member.extension.serialize(),
            }
            for member in self.members
        ]
        return data


async def initialize_database(connection, stage2_only=False, stage1_only=False):
    if not stage2_only:
        await connection.execute(
            "CREATE TYPE extension_type AS ENUM('SIMPLE', 'MULTIRING', 'GROUP', 'EXTERNAL', "
            "'TRUNK')"
        )
        await connection.execute(
            "CREATE TYPE forwarding_mode AS ENUM('DISABLED', 'ENABLED', 'ON_BUSY', "
            "'ON_UNAVAILABLE')"
        )
        await connection.execute(
            "CREATE TYPE fork_rank_mode AS ENUM('DEFAULT', 'NEXT', 'DROP')"
        )
        await connection.execute(
            "CREATE TYPE fork_rankmember_type AS ENUM('DEFAULT', 'AUXILIARY', 'PERSISTENT')"
        )

    if not stage1_only:
        await connection.execute(
            "CREATE TYPE dect_displaymode AS ENUM('NUMBER', 'NUMBER_AND_NAME', 'NAME')"
        )

    if not stage2_only:
        await connection.execute(CreateTable(Yate.table))
        await connection.execute(CreateTable(Extension.table))
        await connection.execute(CreateTable(ForkRank.table))
        await connection.execute(CreateTable(ForkRank.member_table))

    if not stage1_only:
        await connection.execute(CreateTable(User.table))
        await connection.execute(CreateTable(Registration.table))
        await connection.execute(CreateTable(ActiveCall.table))


async def regenerate_database_objects(connection, stage2_only=False, stage1_only=False):
    if not stage2_only:
        await connection.execute('DROP TABLE IF EXISTS "Yate" CASCADE')
        await connection.execute('DROP TABLE IF EXISTS "Extension" CASCADE')
        await connection.execute('DROP TABLE IF EXISTS "ForkRank" CASCADE')
        await connection.execute('DROP TABLE IF EXISTS "ForkRankMember" CASCADE')

    if not stage1_only:
        await connection.execute("DROP TABLE IF EXISTS users CASCADE")
        await connection.execute("DROP TABLE IF EXISTS active_calls CASCADE")
        await connection.execute("DROP TABLE IF EXISTS registrations CASCADE")

        await connection.execute("DROP TYPE IF EXISTS dect_displaymode")

    if not stage2_only:
        await connection.execute("DROP TYPE IF EXISTS extension_type")
        await connection.execute("DROP TYPE IF EXISTS forwarding_mode")
        await connection.execute("DROP TYPE IF EXISTS fork_rank_mode")
        await connection.execute("DROP TYPE IF EXISTS fork_rankmember_type")

    await initialize_database(connection, stage2_only, stage1_only)


def main():
    asyncio.run(amain())


async def amain():
    # init database
    import argparse
    import sys

    from aiopg.sa import create_engine

    import ywsd.settings

    parser = argparse.ArgumentParser(description="Yate Routing Engine")
    parser.add_argument(
        "--config", type=str, help="Config file to use.", default="routing_engine.yaml"
    )
    parser.add_argument(
        "--stage2", help="Only setup tables for stage2 routing", action="store_true"
    )
    parser.add_argument(
        "--stage1", help="Only setup tables for stage1 routing", action="store_true"
    )
    parser.add_argument(
        "--regenerate", help="Drop tables if they already exist", action="store_true"
    )

    args = parser.parse_args()
    settings = ywsd.settings.Settings(args.config)

    async with create_engine(**settings.DB_CONFIG) as engine:
        async with engine.acquire() as conn:
            if args.regenerate:
                await regenerate_database_objects(conn, args.stage2, args.stage1)
            else:
                await initialize_database(conn, args.stage2, args.stage1)


if __name__ == "__main__":
    main()



import asyncio
import functools
import logging
from importlib import import_module

from psycopg2 import OperationalError


def class_from_dotted_string(dotted_string):
    try:
        module_name, class_name = dotted_string.rsplit(".", 1)
    except ValueError as e:
        raise ImportError("Invalid dotted string: {}".format(dotted_string))

    module = import_module(module_name)

    try:
        return getattr(module, class_name)
    except AttributeError as e:
        raise ImportError(
            "There is no class {} in module {}".format(class_name, module_name)
        )


def retry_db_offline(count, wait_ms):
    def decorate(function):
        @functools.wraps(function)
        async def decorated(*args, **kwargs):
            for _ in range(count):
                try:
                    return await function(*args, **kwargs)
                except OperationalError as e:
                    logging.warning("Database error: {}. Waiting to retry...".format(e))
                await asyncio.sleep(wait_ms / 1000)
            logging.error("Continued database error. Stopped retrying....")

        return decorated

    return decorate



import logging
from enum import Enum
from typing import List, Dict, Tuple, Optional
import os.path
import uuid

from ywsd.objects import Extension, ForkRank, Yate, DoesNotExist


class RoutingError(Exception):
    def __init__(self, error_code, message):
        self.error_code = error_code
        self.message = message


class RoutingTree:
    def __init__(self, source, target, source_params, settings):
        self.source_extension = source
        self.target_extension = target
        self.source = None
        self.target = None
        self._settings = settings
        self._source_params = source_params

        self.routing_result = None
        self.new_routing_cache_content = {}
        self.all_routing_results = {}

    async def discover_tree(
        self, db_connection
    ) -> Optional["RoutingTreeDiscoveryVisitor"]:
        await self._load_source_and_target(db_connection)
        if self.target.type != Extension.Type.TRUNK:
            visitor = RoutingTreeDiscoveryVisitor(self.target, [self.source.extension])
            await visitor.discover_tree(db_connection)
            return visitor

    def calculate_routing(
        self, local_yate: int, yates_dict: Dict[int, Yate]
    ) -> Tuple["IntermediateRoutingResult", Dict[str, "IntermediateRoutingResult"]]:
        self._calculate_main_routing(local_yate, yates_dict)
        self._provide_ringback()
        self._populate_source_and_target_parameters()

        return self.routing_result, self.new_routing_cache_content

    def serialized_tree(self):
        return self.target.serialize() if self.target else None

    def _calculate_main_routing(self, local_yate, yates_dict):
        visitor = YateRoutingGenerationVisitor(self, local_yate, yates_dict)
        if self.target.type != Extension.Type.TRUNK:
            result = visitor.calculate_routing()
        else:
            result = visitor.generate_trunk_routing(self.target, self.target_extension)
        self.routing_result = result
        self.all_routing_results = visitor.get_routing_results()
        self.new_routing_cache_content = visitor.get_routing_cache_content()
        if not result.is_valid:
            raise RoutingError("noroute", "The main routing target returned NO_ROUTE.")

    def _provide_ringback(self):
        if self.target.ringback is not None:
            ringback_path = (
                os.path.join(
                    self._settings.RINGBACK_TOP_DIRECTORY, self.target.ringback
                )
                + ".slin"
            )
            if os.path.isfile(ringback_path):
                if self.routing_result.is_simple:
                    # we need to convert routing result into a simple fork
                    self.routing_result = IntermediateRoutingResult(
                        target=CallTarget(
                            "fork", self.routing_result.target.parameters
                        ),
                        fork_targets=[
                            self._make_ringback_target(ringback_path),
                            self.routing_result.target,
                        ],
                    )
                else:
                    # if the routing target is already a callfork, just prepend the ringback target to the first group
                    self.routing_result.fork_targets.insert(
                        0, self._make_ringback_target(ringback_path)
                    )

    def _populate_parameters_global(self, parameters):
        self.routing_result.target.parameters.update(parameters)
        for entry in self.new_routing_cache_content.values():
            entry.target.parameters.update(parameters)

    def _populate_source_and_target_parameters(self):
        parameters = self._source_params
        if self.target.name is not None:
            parameters["calledname"] = self.target.name

        if self.target.short_name is not None:
            # Callername should always be populated by source parameters, otherwise, default to source name
            callername = self._source_params.get("callername", self.source.name)
            parameters["callername"] = "[{}] {}".format(
                self.target.short_name, callername
            )

        self._populate_parameters_global(parameters)

    @staticmethod
    def _make_ringback_target(path):
        return CallTarget(
            "wave/play/" + path,
            {
                "fork.calltype": "persistent",
                "fork.autoring": "true",
                "fork.automessage": "call.progress",
            },
        )

    async def _load_source_and_target(self, db_connection):
        try:
            if isinstance(self.source_extension, Extension):
                self.source = self.source_extension
            else:
                self.source = await Extension.load_extension(
                    self.source_extension, db_connection
                )
        except DoesNotExist:
            self.source = Extension.create_unknown(self.source_extension)
        try:
            self.target = await Extension.load_extension(
                self.target_extension, db_connection
            )
            self.target.tree_identifier = str(self.target.id)
        except DoesNotExist:
            # we give this one rescue attempt by trying to load a trunk
            try:
                self.target = await Extension.load_trunk_extension(
                    self.target_extension, db_connection
                )
                self.target.tree_identifier = str(self.target.id)
                return
            except DoesNotExist:
                pass
            raise RoutingError("noroute", "Routing target was not found")


class RoutingTreeDiscoveryVisitor:
    def __init__(self, root_node, excluded_targets, max_depth=10):
        self._root_node = root_node
        self._excluded_targets = set(excluded_targets)
        self._max_depth = max_depth
        self._failed = False
        self._pruned = False

    @property
    def failed(self):
        return self._failed

    @property
    def pruned(self):
        return self._pruned

    async def discover_tree(self, db_connection):
        await self._visit(
            self._root_node, 0, list(self._excluded_targets), db_connection
        )

    async def _visit(
        self, node: Extension, depth: int, path_extensions: list, db_connection
    ):
        if depth >= self._max_depth:
            node.routing_log(
                "Routing aborted due to depth limit at {}".format(node), "ERROR"
            )
            self._failed = True
            return

        path_extensions_local = path_extensions.copy()
        path_extensions_local.append(node.extension)

        if (
            node.type != Extension.Type.EXTERNAL
            and node.forwarding_mode != Extension.ForwardingMode.DISABLED
        ):
            await node.load_forwarding_extension(db_connection)
        if node.type in (Extension.Type.GROUP, Extension.Type.MULTIRING) and (
            node.forwarding_mode != Extension.ForwardingMode.ENABLED
            or node.forwarding_delay > 0
        ):
            # we discover group members if there is no immediate forward
            await node.populate_fork_ranks(db_connection)
        # now we visit the populated children if they haven't been already discovered
        if node.forwarding_extension is not None:
            # TODO: We might want to avoid following forwards if this is discovered as a MULTIRING child?
            fwd = node.forwarding_extension
            if fwd.extension not in path_extensions_local:
                await self._visit(fwd, depth + 1, path_extensions_local, db_connection)
            else:
                self._pruned = True
                node.routing_log(
                    "Discovery aborted for forward to {}, was already present.\n"
                    "Disabling Forward".format(fwd),
                    "WARN",
                    related_node=node.forwarding_extension,
                )
                node.forwarding_mode = Extension.ForwardingMode.DISABLED
        for fork_rank in node.fork_ranks:
            for member in fork_rank.members:
                # do not discover inactive members
                if not member.active:
                    continue
                ext = member.extension
                if ext.extension not in path_extensions_local:
                    await self._visit(
                        ext, depth + 1, path_extensions_local, db_connection
                    )
                else:
                    self._pruned = True
                    fork_rank.routing_log(
                        "Discovery aborted for {} in {}, was already present.\n"
                        "Temporarily disable membership for this routing.".format(
                            ext, fork_rank, path_extensions_local
                        ),
                        "WARN",
                        related_node=member.extension,
                    )
                    member.active = False


class CallTarget:
    target = ""
    parameters = {}

    def __init__(self, target, parameters=None):
        self.target = target
        self.parameters = parameters if parameters is not None else {}

    def serialize(self):
        return {"target": self.target, "parameters": self.parameters}

    @classmethod
    def deserialize(cls, data):
        target = data["target"]
        parameters = data.get("parameters")
        return cls(target, parameters=parameters)

    @property
    def is_separator(self):
        return self.target.startswith("|")

    def __repr__(self):
        return "<CallTarget {}, params={}>".format(self.target, self.parameters)


class IntermediateRoutingResult:
    class Type(Enum):
        SIMPLE = 0
        FORK = 1
        NO_ROUTE = 99

    def __init__(
        self, target: CallTarget = None, fork_targets: List["CallTarget"] = None
    ):
        if fork_targets is not None:
            if len(fork_targets) > 0:
                self.type = IntermediateRoutingResult.Type.FORK
                self.fork_targets = fork_targets
                self.target = target
            else:
                self.type = IntermediateRoutingResult.Type.NO_ROUTE
                self.target = None
                self.fork_targets = []
        elif target is not None:
            self.type = IntermediateRoutingResult.Type.SIMPLE
            self.target = target
            self.fork_targets = []
        else:
            self.type = IntermediateRoutingResult.Type.NO_ROUTE
            self.target = None
            self.fork_targets = []

    def serialize(self):
        result = {
            "type": str(self.type),
            "target": self.target.serialize() if self.target is not None else str(None),
        }
        if self.fork_targets:
            result["fork_targets"] = [
                target.serialize() for target in self.fork_targets
            ]
        return result

    @classmethod
    def deserialize(cls, data):
        target = data.get("target")
        if target is not None:
            target = CallTarget.deserialize(target)
        fork_targets = [
            CallTarget.deserialize(target) for target in data.get("fork_targets", [])
        ]
        return cls(target=target, fork_targets=fork_targets)

    def __repr__(self):
        fork_targets_str = "\n\t\t".join([repr(targ) for targ in self.fork_targets])
        return "<IntermediateRoutingResult\n\ttarget={}\n\tfork_targets=\n\t\t{}\n>".format(
            self.target, fork_targets_str
        )

    @property
    def is_simple(self):
        return self.type == IntermediateRoutingResult.Type.SIMPLE

    @property
    def is_valid(self):
        return self.type != IntermediateRoutingResult.Type.NO_ROUTE


class YateRoutingGenerationVisitor:
    def __init__(
        self, routing_tree: RoutingTree, local_yate_id: int, yates_dict: Dict[int, Yate]
    ):
        self._routing_tree = routing_tree
        self._local_yate_id = local_yate_id
        self._yates_dict = yates_dict
        self._lateroute_cache: Dict[str, IntermediateRoutingResult] = {}
        self._x_eventphone_id = uuid.uuid4().hex
        self._routing_results: Dict[str, IntermediateRoutingResult] = {}

    def get_routing_cache_content(self):
        return self._lateroute_cache

    def get_routing_results(self):
        return self._routing_results

    def _make_intermediate_result(self, target: CallTarget = None, fork_targets=None):
        return IntermediateRoutingResult(target=target, fork_targets=fork_targets)

    def _make_calltarget(self, target: str, parameters: dict = None):
        # write default parameters into the calltarget
        if parameters is None:
            parameters = {}
        parameters["x_eventphone_id"] = self._x_eventphone_id
        parameters["osip_X-Eventphone-Id"] = self._x_eventphone_id
        return CallTarget(target=target, parameters=parameters)

    def _cache_intermediate_result(self, result: IntermediateRoutingResult):
        if not result.is_simple:
            self._lateroute_cache[result.target.target] = result

    def calculate_routing(self):
        return self._visit(self._routing_tree.target, [])

    def _visit(self, node: Extension, path: list) -> IntermediateRoutingResult:
        result = self._visit_for_route_calculation(node, path)
        self._routing_results[node.tree_identifier] = result
        return result

    def _visit_for_route_calculation(
        self, node: Extension, path: list
    ) -> IntermediateRoutingResult:
        local_path = path.copy()
        local_path.append(node.id)

        # first we check if this node has an immediate forward. If yes, we defer routing there.
        if node.immediate_forward:
            return self._visit(node.forwarding_extension, local_path)

        if YateRoutingGenerationVisitor.node_has_simple_routing(node):
            return self._make_intermediate_result(
                target=self.generate_simple_routing_target(node)
            )
        else:
            # this will require a fork
            # go through the callgroup ranks to issue the groups of the fork
            fork_targets = []
            accumulated_delay = 0
            for rank in node.fork_ranks:
                if fork_targets:
                    # this is not the first rank, so we need to generate a separator
                    if rank.mode == ForkRank.Mode.DROP:
                        separator = "|drop={}".format(rank.delay)
                        accumulated_delay += rank.delay
                    elif rank.mode == ForkRank.Mode.NEXT:
                        separator = "|next={}".format(rank.delay)
                        accumulated_delay += rank.delay
                    else:
                        separator = "|"
                        # If we see an untimed separator, any time-based forward is not possible anymore
                        if node.forwarding_mode == Extension.ForwardingMode.ENABLED:
                            node.routing_log(
                                "Non time-based fork rank is incompatible with time-based forward. "
                                "Disabling the forward.",
                                "WARN",
                                rank,
                            )
                            node.forwarding_mode = Extension.ForwardingMode.DISABLED

                    if (
                        node.forwarding_mode == Extension.ForwardingMode.ENABLED
                        and accumulated_delay >= node.forwarding_delay
                    ):
                        # all of those will not be called, as the forward takes effect now
                        node.routing_log(
                            "Fork rank (and following) are ignored due to time-based forward.",
                            "WARN",
                            rank,
                        )
                        break
                    # Do not generate default params on pseudo targets
                    fork_targets.append(CallTarget(separator))
                for member in rank.members:
                    # do not route inactive members
                    if not member.active:
                        continue
                    member_route = self._visit(member.extension, local_path)
                    if not member_route.is_valid:
                        rank.routing_log(
                            "Extension has no valid (non-empty) routing and is thus ignored.",
                            "WARN",
                            member.extension,
                        )
                        continue
                    if member.type.is_special_calltype:
                        member_route.target.parameters[
                            "fork.calltype"
                        ] = member.type.fork_calltype
                    # please note that we ignore the member modes for the time being
                    fork_targets.append(member_route.target)
                    self._cache_intermediate_result(member_route)

                if fork_targets and fork_targets[-1].target == "|":
                    # We just created an empty default rank. This will cause the call to hang
                    del fork_targets[-1]
                    rank.routing_log(
                        "This created an empty default rank. It will be removed to prevent call hang.",
                        "WARN",
                    )

            # if this is a MULTIRING or (SIMPLE with forward), the extension itself needs to be part of the first group
            if node.type in (Extension.Type.MULTIRING, Extension.Type.SIMPLE):
                # in difference to groups, the first ForkRank can have type NEXT or DROP and we should respect it
                if len(node.fork_ranks) > 0:
                    first_fork_rank = node.fork_ranks[0]
                    if first_fork_rank.mode == ForkRank.Mode.NEXT:
                        fork_targets.insert(
                            0, CallTarget("|next={}".format(first_fork_rank.delay))
                        )
                    elif first_fork_rank.mode == ForkRank.Mode.DROP:
                        fork_targets.insert(
                            0, CallTarget("|drop={}".format(first_fork_rank.delay))
                        )
                    # If the fork rank is default, we assume that multiring should start with the main extension
                    # and do nothing here
                fork_targets.insert(0, self.generate_simple_routing_target(node))

            # Handle forwards
            if node.forwarding_mode == Extension.ForwardingMode.ON_BUSY:
                # There should be no call waiting on all previous call legs.
                for target in fork_targets:
                    if not target.is_separator:
                        target.parameters["osip_X-No-Call-Wait"] = "1"

            if node.forwarding_mode in (
                Extension.ForwardingMode.ENABLED,
                Extension.ForwardingMode.ON_BUSY,
                Extension.ForwardingMode.ON_UNAVAILABLE,
            ):
                # this is non-immediate forward
                forwarding_route = self._visit(node.forwarding_extension, local_path)
                if node.forwarding_mode == Extension.ForwardingMode.ENABLED:
                    fwd_delay = node.forwarding_delay - accumulated_delay
                    fork_targets.append(CallTarget("|drop={}".format(fwd_delay)))
                else:
                    # Add a default rank, call will progress to next rang when all previous calls failed
                    fork_targets.append(CallTarget("|"))
                fork_targets.append(forwarding_route.target)
                self._cache_intermediate_result(forwarding_route)

            return self._make_intermediate_result(
                fork_targets=fork_targets,
                target=self._make_calltarget(
                    self.generate_deferred_routestring(local_path)
                ),
            )

    @staticmethod
    def node_has_simple_routing(node: Extension):
        if node.type == Extension.Type.EXTERNAL:
            return True
        # if the node is set to immediate forward, it will be ignored for routing and we calculate the routing for
        # the forwarded node instead
        if node.immediate_forward:
            return YateRoutingGenerationVisitor.node_has_simple_routing(
                node.forwarding_extension
            )

        if node.type == Extension.Type.SIMPLE:
            if node.forwarding_mode == Extension.ForwardingMode.DISABLED:
                return True
            # Forwarding with delay or ON_BUSY/UNAVAILABLE requires a callfork
            return False
        # multiring is simple, if there are no active multiring participants configured
        if node.type == Extension.Type.MULTIRING:
            if node.has_active_group_members:
                return False
            # ok, nothing to multiring, so look at the forwarding mode
            if node.forwarding_mode == Extension.ForwardingMode.DISABLED:
                return True
            # Forwarding with delay or ON_BUSY/UNAVAILABLE requires a callfork
            return False
        # groups might have a simple routing if they have exactly one participant. We will ignore this possibility
        # for the moment being. We could do this by introducing an optimizer stage that reshapes the tree :P
        return False

    def generate_simple_routing_target(self, node: Extension):
        if node.type == Extension.Type.EXTERNAL:
            # External things are handled by regexroute, just issue a lateroute that triggers this
            return self._make_calltarget(
                "lateroute/{}".format(node.extension), {"eventphone_stage2": "1"}
            )
        if node.yate_id is None:
            raise RoutingError(
                "failure",
                "Extension {} is misconfigured - yate_id is NULL.".format(node),
            )
        if node.yate_id == self._local_yate_id:
            return self._make_calltarget(
                "lateroute/{}".format(node.extension), {"eventphone_stage2": "1"}
            )
        else:
            return self._make_calltarget(
                "sip/sip:{}@{}".format(
                    node.extension, self._yates_dict[node.yate_id].hostname
                ),
                {
                    "oconnection_id": self._yates_dict[node.yate_id].voip_listener,
                },
            )

    def generate_trunk_routing(self, trunk: Extension, dialed_number: str):
        if trunk.yate_id is None:
            raise RoutingError(
                "failure",
                "Trunk-Extension {} is misconfigured - yate_id is NULL.".format(trunk),
            )
        if trunk.yate_id == self._local_yate_id:
            return IntermediateRoutingResult(
                target=self._make_calltarget(
                    "lateroute/{}".format(dialed_number), {"eventphone_stage2": "1"}
                )
            )
        else:
            return IntermediateRoutingResult(
                target=self._make_calltarget(
                    "sip/sip:{}@{}".format(
                        dialed_number, self._yates_dict[trunk.yate_id].hostname
                    ),
                    {
                        "oconnection_id": self._yates_dict[trunk.yate_id].voip_listener,
                    },
                )
            )

    def generate_deferred_routestring(self, path):
        return "lateroute/" + self.generate_node_route_string(path)

    def generate_node_route_string(self, path):
        joined_path = "-".join(map(str, path))
        return "stage1-{}-{}".format(self._x_eventphone_id, joined_path)



import yaml


class Settings:
    def __init__(self, config_file=None):
        if config_file is None:
            config_file = "routing_engine.yaml"
        with open(config_file, "r") as f:
            self.config = yaml.safe_load(f)

    def __getattr__(self, item):
        return self.config.get(item)



import logging
from typing import List

from yate.protocol import Message

from ywsd.objects import User, ActiveCall, Registration, DoesNotExist
from ywsd.util import retry_db_offline


HEADER_NAMES = (
    "X-Eventphone-Id",
    "X-No-Call-Wait",
)


def get_headers(msg: Message):
    headers = {}
    for header in HEADER_NAMES:
        value = msg.params.get("osip_" + header)
        if value is None:
            value = msg.params.get("sip_" + header.lower())
        headers[header] = value
    return headers


class RoutingTask:
    def __init__(self, yate: "ywsd.engine.YateRoutingEngine", message: Message):
        self._yate = yate
        self._message = message

    def populate_additional_message_parameters(self, headers):
        self._message.params["X-Eventphone-Id"] = headers.get("X-Eventphone-Id", "")

        # Make cdrbuild import the X-Eventphone-Id into the cdr record so that we can grab it from call.cdr
        if "copyparams" in self._message.params:
            self._message.params["copyparams"] += ",X-Eventphone-Id"
        else:
            self._message.params["copyparams"] = "X-Eventphone-Id"

    async def _calculate_stage2_routing(self, caller, called):
        if called.startswith("stage2-"):
            called = called[7:]

        async with self._yate.stage2_db_engine.acquire() as db_connection:
            try:
                target = await User.load_user(called, db_connection)
            except DoesNotExist:
                try:
                    target = await User.load_trunk(called, db_connection)
                except DoesNotExist:
                    return False, False

            if target.type == "static":
                return self._static_target_routing(target)

            locations = await Registration.load_locations_for(
                target, called, db_connection
            )
            if not locations:
                self._message.params["error"] = "offline"
                self._message.params["reason"] = "offline"
                return False, True

            headers = get_headers(self._message)

            # Check if this call should be dropped
            if (
                headers["X-No-Call-Wait"] == "1" or not target.call_waiting
            ) and target.inuse > 0:
                self._message.params["error"] = "busy"
                return False, True
            if await ActiveCall.is_active_call(
                "called", headers["X-Eventphone-Id"], db_connection
            ):
                self._message.params["error"] = "busy"
                return False, True

            # calculate target(s)
            if len(locations) == 1:
                self._message.return_value = locations[0].call_target
                self._message.params["oconnection_id"] = locations[0].oconnection_id
            else:
                self._message.return_value = "fork"
                for i, location in enumerate(locations, start=1):
                    self._message.params["callto.{}".format(i)] = location.call_target
                    self._message.params[
                        "callto.{}.oconnection_id".format(i)
                    ] = location.oconnection_id

            self.populate_additional_message_parameters(headers)
            return True, True

    def _static_target_routing(self, target):
        try:
            separated_target = target.static_target.split(";")
            message_params = self._process_static_target_parameters(
                separated_target[1:]
            )
        except ValueError:
            logging.error(
                f"Encountered invalid static call target:'{target.static_target}'"
            )
            self._message.params["error"] = "failure"
            return False, True
        self._message.return_value = separated_target[0]
        headers = get_headers(self._message)
        self.populate_additional_message_parameters(headers)
        self._message.params.update(message_params)
        return True, True

    @staticmethod
    def _process_static_target_parameters(params: List[str]):
        result = {}
        for param in params:
            key, value = param.split("=", 1)
            result[key] = value
        return result

    @retry_db_offline(count=4, wait_ms=1000)
    async def routing_job(self):
        caller = self._message.params.get("caller")
        called = self._message.params.get("called")

        logging.debug("Doing stage2 routing from {} to {}".format(caller, called))

        if caller is None:
            # we do not process messages without a caller
            self._yate.answer_message(self._message, False)
            return

        success, handled = await self._calculate_stage2_routing(caller, called)
        if success:
            logging.debug(
                "Routing successful. Target is {}".format(self._message.return_value)
            )
        elif handled:
            logging.debug(
                "Routing not successful. Error is {}.".format(
                    self._message.params["error"]
                )
            )
        else:
            logging.debug("Routing not successful, noroute, pass message on")

        self._yate.answer_message(self._message, handled)



from setuptools import setup
import os

this_dir = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_dir, "README.md"), "r") as f:
    long_description = f.read()

setup(
    name="ywsd",
    version="0.10.0",
    packages=["ywsd"],
    url="https://gitlab.rc5.de/eventphone/ywsd",
    license="AGPLv3+",
    author="Garwin (Martin Lang)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email="garwin@eventphone.de",
    description="A yate routing engine for event telephone networks",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
    ],
    install_requires=[
        "aiopg",
        "aiohttp",
        "python-yate>=0.3",
        "pyyaml",
        "sqlalchemy",
    ],
    entry_points={
        "console_scripts": [
            "ywsd_init_db=ywsd.objects:main",
            "ywsd_engine=ywsd.engine:main",
        ],
    },
)
